'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.optionsFromOData = exports.testForOData = undefined;

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _odataParser = require('odata-parser');

var _odataParser2 = _interopRequireDefault(_odataParser);

var _url = require('url');

var _url2 = _interopRequireDefault(_url);

var _nodeBits = require('node-bits');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; } /* eslint-disable no-use-before-define */


var ODATA_FLAG = '$';
var ESCAPE_AND_HOPE_WE_NEVER_SEE = 'JaQzEpW';

// helpers
var buildColumnFuncCompare = function buildColumnFuncCompare(left, comparator) {
  if (left.type && left.type === 'functioncall') {
    var map = oDataFuncMap[left.func];
    return { columnFunc: map(left), compare: comparator };
  }
  return _defineProperty({}, parseNode(left), comparator);
};

var oDataFuncMap = {
  substringof: function substringof(node) {
    return buildColumnFuncCompare(node.args[0], { like: parseNode(node.args[1]) });
  }, // this maps to contains
  startswith: function startswith(node) {
    return buildColumnFuncCompare(node.args[0], { startsWith: parseNode(node.args[1]) });
  },
  endswith: function endswith(node) {
    return buildColumnFuncCompare(node.args[0], { endsWith: parseNode(node.args[1]) });
  },
  tolower: function tolower(node) {
    return { func: 'tolower', args: node.args };
  }
};

var oDataTypeMap = {
  eq: function eq(node) {
    return _defineProperty({}, parseNode(node.left), parseNode(node.right));
  },
  ne: function ne(node) {
    return _defineProperty({}, parseNode(node.left), { ne: parseNode(node.right) });
  },
  gt: function gt(node) {
    return _defineProperty({}, parseNode(node.left), { gt: parseNode(node.right) });
  },
  ge: function ge(node) {
    return _defineProperty({}, parseNode(node.left), { ge: parseNode(node.right) });
  },
  lt: function lt(node) {
    return _defineProperty({}, parseNode(node.left), { lt: parseNode(node.right) });
  },
  le: function le(node) {
    return _defineProperty({}, parseNode(node.left), { le: parseNode(node.right) });
  },
  and: function and(node) {
    return { and: [parseNode(node.left), parseNode(node.right)] };
  },
  or: function or(node) {
    return { or: [parseNode(node.left), parseNode(node.right)] };
  },
  functioncall: function functioncall(node) {
    return parseFunctionCall(node);
  },
  property: function property(node) {
    return node.name;
  },
  literal: function literal(node) {
    return node.value;
  }
};

var parseFunctionCall = function parseFunctionCall(root) {
  var map = oDataFuncMap[root.func];
  if (!map) {
    throw new Error('Unsupported function call - ' + root.func);
  }

  return map(root);
};

var parseColumnFunction = function parseColumnFunction(root) {
  var map = oDataFuncMap[root.left.func];
  if (!map) {
    throw new Error('Unsupported function call - ' + root.left.func);
  }

  return { columnFunc: map(root.left), compare: _defineProperty({}, root.type, parseNode(root.right)) };
};

var parseNode = function parseNode(root) {
  var map = root.left && root.left.type === 'functioncall' ? parseColumnFunction : oDataTypeMap[root.type];
  if (!map) {
    throw new Error('Unsupported node type - ' + root.type);
  }

  return map(root);
};

var buildWhereClause = function buildWhereClause(oDataFilter) {
  if (!oDataFilter) {
    return undefined; // eslint-disable-line
  }

  return parseNode(oDataFilter);
};

var buildOrderByClause = function buildOrderByClause(oDataOrderBy) {
  if (!oDataOrderBy) {
    return undefined; // eslint-disable-line
  }

  return oDataOrderBy.map(function (item) {
    var key = _lodash2.default.keys(item)[0];
    return {
      field: key,
      direction: item[key]
    };
  });
};

var hackQueryString = function hackQueryString(urlstring) {
  var querystring = decodeURI(_url2.default.parse(urlstring).query);

  // replace + with spaces
  querystring = querystring.replace(/\+/g, ' ');

  // handle 4.0 replacement of substring with contains, the node-odata-parser doesnt currently work for this
  querystring = querystring.replace(/contains/g, 'substringof');

  // handle support of . and / for traversing the stack
  querystring = querystring.replace(/\./g, ESCAPE_AND_HOPE_WE_NEVER_SEE);
  querystring = querystring.replace(/\//g, ESCAPE_AND_HOPE_WE_NEVER_SEE);

  return {
    string: querystring
  };
};

var escapeString = function escapeString(string) {
  return string.replace(new RegExp(ESCAPE_AND_HOPE_WE_NEVER_SEE, 'g'), '.');
};

var replacePeriodEscape = function replacePeriodEscape(node) {
  if (!node) {
    return;
  }

  // if its an array, we need to evaluate each child
  if (_lodash2.default.isArray(node)) {
    _lodash2.default.forEach(node, function (child, index) {
      if (_lodash2.default.isString(child)) {
        node.splice(index, 1, escapeString(child));
        return;
      }

      if (_lodash2.default.isObject(child)) {
        replacePeriodEscape(child);
      }
    });

    return;
  }

  // object, evaluate each key
  if (_lodash2.default.isObject(node)) {
    _lodash2.default.forOwn(node, function (value, key) {
      var newKey = key;
      if (key.includes(ESCAPE_AND_HOPE_WE_NEVER_SEE)) {
        delete node[key];
        newKey = key.replace(new RegExp(ESCAPE_AND_HOPE_WE_NEVER_SEE, 'g'), '.');
      }

      if (_lodash2.default.isString(value)) {
        node[newKey] = escapeString(value);
      } else {
        node[newKey] = value;
        replacePeriodEscape(value);
      }
    });
  }
};

var unhackQueryString = function unhackQueryString(query, oDataQuery) {
  replacePeriodEscape(oDataQuery.$select);
  replacePeriodEscape(oDataQuery.$orderby);
  replacePeriodEscape(oDataQuery.$filter);
  replacePeriodEscape(oDataQuery.$expand);

  return oDataQuery;
};

// exports
var testForOData = exports.testForOData = function testForOData(req) {
  return _lodash2.default.some(req.query, function (value, key) {
    return key.startsWith(ODATA_FLAG);
  });
};

var optionsFromOData = exports.optionsFromOData = function optionsFromOData(req) {
  var query = hackQueryString(req.url);
  if (!query) {
    return {};
  }

  var rawODataQuery = _odataParser2.default.parse(unescape(query.string));
  var oDataQuery = unhackQueryString(query, rawODataQuery);

  return {
    includeMetaData: [{ key: '@odata.count', value: _nodeBits.COUNT }],

    start: oDataQuery.$skip || 0,
    max: oDataQuery.$top || undefined, // eslint-disable-line
    select: oDataQuery.$select || undefined, // eslint-disable-line
    orderby: buildOrderByClause(oDataQuery.$orderby),
    where: buildWhereClause(oDataQuery.$filter),
    expand: oDataQuery.$expand
  };
};